#
#  Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015
#            Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Utility routines for Crates.

Changes in CIAO 4.6 include:

 - the add_keyval routine has been removed since the set_key
   routine from the Crates module should be used instead.

 - SimpleCoordTransform now supports the pytransform.WCSTransform
   class (the 'replacement' for pytransform.WCSTANTransform)

"""

import os
import time
import collections

import numpy as np
import pycrates

from ciao_contrib.logger_wrapper import initialize_module_logger
import ciao_contrib.smooth as sm

__all__ = (
    "make_table_crate", "make_image_crate",
    "add_colvals",
    "write_columns", "write_arrays",
    "scale_image_crate",
    "smooth_image_crate",
    "SimpleCoordTransform",
    "read_ds9_contours",
)

logger = initialize_module_logger("crates_utils")

v3 = logger.verbose3

# time.strftime format to create FITS (like) time format
_fits_time_fmt = "%Y-%m-%dT%H:%M:%S"


def add_colvals(cr, colname, colvals, unit=None, desc=None):
    """Create a new column in the crate with the given
    name and data values. The unit and desc fields are
    used to set the column unit and description if given.

    You can not over-write an existing column (a ValueError
    is raised in this case).

    """

    # In CIAO 4.4 there is some confusion within Crates if you
    # over-write a column, so force users to be explicit here.
    if cr.column_exists(colname):
        raise ValueError("Column {} already exists in the table crate.".format(colname))

    cd = pycrates.CrateData()
    cd.name = colname

    cd.values = colvals
    if unit is not None:
        cd.unit = unit
    if desc is not None:
        cd.desc = desc

    cr.add_column(cd)


def _sarray_to_crate(sa, colnames=None):
    """Create a TABLECrate that contains the contents
    of the supplied structured/rec array. If colnames is
    not None then it it an array of names that determines
    the order to add the columns to the crate (and can
    be used to filter out columns in sa).

    The arrays in the structured array are explicitly
    copied into the crate to avoid some memory issues
    seen in CIAO 4.4 testing. However, this may not
    resolve all issues so it is suggested that this
    routine is not used at this time.

    Not all data types are necessarily supported.
    """

    cr = pycrates.TABLECrate()
    cr.name = "TABLE"

    if colnames is None:
        names = sa.dtype.names
    else:
        names = colnames

    for cname in names:
        # We copy the values to avoid an issue on writing out the
        # crate; not sure if this is a Crates or NumPy problem
        # and whether this has been fixed since CIAO 4.4 when it
        # was seen
        #
        add_colvals(cr, cname, sa[cname].copy())

    return cr


def _dict_to_crate(d, colnames=None):
    """Create a TABLECrate that contains the contents of the
    dictionary. If colnames is not None then it it an array of names
    that determines the order to add the columns to the crate (and can
    be used to filter out columns). Use of collections.OrderedDict
    rather than a normal dictionary can also be used to determine
    the order of the columns.

    The data arrays used in the crate are references of the
    input values (i.e. they are not copied, unlike _sarray_to_crate).

    Not all data types are necessarily supported.
    """

    cr = pycrates.TABLECrate()
    cr.name = "TABLE"

    if colnames is None:
        names = d.keys()
    else:
        names = colnames

    for cname in names:
        add_colvals(cr, cname, d[cname])

    return cr


def _arrays_to_crate(arrays, colnames=None):
    """Create a TABLECrate that contains the contents of the
    columns in arrays (each element is a column).

    If colnames is not None then they are used as the column names,
    otherwise values of col1 to colN are used.

    The data arrays used in the crate are references of the
    input values (i.e. they are not copied, unlike _sarray_to_crate).

    Not all data types are necessarily supported.
    """

    na = len(arrays)
    if colnames is None:
        colnames = ["col{0}".format(i) for i in range(1, na + 1)]

    else:
        nc = len(colnames)
        if nc != na:
            raise ValueError("Number of columns ({}) and column names ({}) does not match.".format(nc, na))

    cr = pycrates.TABLECrate()
    cr.name = "TABLE"
    for (col, name) in zip(arrays, colnames):
        add_colvals(cr, name, col)

    return cr


def make_table_crate(*args, **kwargs):
    """Create a table Crate from the given columns.

    There is only one keyword argument - colnames - which
    defaults to None.

    If called with one argument and it acts as a dictionary then
    the keys of the value are taken to be the column names and
    the values are the columns. Use a collections.OrderedDict
    to ensure the column order, since it's taken to be the
    key order returned by iteritems. The order can be over-ridden
    with the colnames argument (this also lets you write out only
    a subset of the keys in the dictionary).

    If a single argument is used that is a stuctured array then
    the column names, order, and data is taken from the array. As
    with dictionaries the colnames argument can be used to re-order
    or subset the data.

    The arguments are taken to be the column data to use. If they are
    not the same length then the smaller arrays will be padded with
    the null value for that column type (0 for numeric columns and
    "IND" or "" for string columns). The arrays can be Python arrays
    or numpy arrays, and include strings. Python arrays should contain
    the same data type for each element; if this does not hold then
    there is no guarantee as to the behavior of this routine.

    When a dictionary or structured array is supplied then the colnames
    argument, if set,  is used to order the columns in the crate.
    Columns in the input argument but not mentioned in colnames are
    excluded.

    Otherwise, the colnames argument is used to name the columnes. If
    left as None then the names "col1" to "coln" will be used (n being
    the number of columns).

    The CREATOR and DATE keywords of the crate are set by this
    routine.

    Examples:

      a = [1,2,3,4,5]
      b = ["aa", "bb", "", "d d", "long string"]
      c = np.arange(10).reshape((5,2))
      cr1 = make_table_crate(a,b)
      cr2 = make_table_crate(a,b,c, colnames=["foo","bar","bax"])

      In this case bax is a vector column

      d = { "foo": [1,2,3], "bar": ["xx", "yy"],
            "bax": np.arange(10).reshape((5,2)) }
      cr3 = make_table_crate(d)
      cr4 = make_table_crate(d, colnames=["foo", "bax"])

      e = np.zeros((2,),dtype=('i4,f4,a10'))
      e[:] = [(1,2.,'Hello'),(2,3.,"World")]
      cr5 = make_table_crate(e)
      cr6 = make_table_crate(e, colnames=['f1', f2'])

    Warnings:

    The semantics of whether the input array(s) are copied or
    just referenced in the crate are currently not specified.
    Changing the original array after the crate has been created
    *may* change the value in the Crate, but this behavior should
    not be taken advantage of.

    """

    isdict = False
    isstruct = False

    nargs = len(args)
    if nargs == 0:
        raise TypeError("make_table_crate() takes at least one argument (0 given)")
    elif nargs == 1:
        isdict = isinstance(args[0], collections.Mapping)
        if hasattr(args[0], "dtype") and hasattr(args[0].dtype, "names"):
            isstruct = args[0].dtype.names is not None

        if isdict and isstruct:
            raise ValueError("Unexpected: argument appears to be a map and a structured/rec array!")

    try:
        colnames = kwargs.pop("colnames")
    except KeyError:
        colnames = None

    if len(kwargs) != 0:
        raise TypeError("make_table_crate() got an unexpected keyword argument '{}'".format(kwargs.keys()[0]))

    if isdict:
        cr = _dict_to_crate(args[0], colnames=colnames)

    elif isstruct:
        cr = _sarray_to_crate(args[0], colnames=colnames)

    else:
        cr = _arrays_to_crate(args, colnames=colnames)

    key1 = pycrates.CrateKey()
    key1.name = "CREATOR"
    key1.value = "make_table_crate"
    key1.desc = "tool that created this output"
    cr.add_key(key1)

    key2 = pycrates.CrateKey()
    key2.name = "DATE"
    key2.value = time.strftime(_fits_time_fmt)
    key2.desc = "Date and time of file creation"
    cr.add_key(key2)

    return cr


def make_image_crate(pixvals):
    """Create an image Crate from the given array.

    The CREATOR and DATE keywords of the crate are set by this
    routine.

    Examples:

      a = np.arange(400*500).reshape(400,500)
      cr = make_image_crate(a)

    Warnings:

    The semantics of whether the input array is copied or
    just referenced in the crate are currently not specified.
    Changing the original array after thecrate has been created
    *may* change the value in the Crate, but this behavior should
    not be taken advantage of.

    """

    cd = pycrates.CrateData()
    cd.values = pixvals

    cr = pycrates.IMAGECrate()
    cr.name = "IMAGE"
    cr.add_image(cd)

    key1 = pycrates.CrateKey()
    key1.name = "CREATOR"
    key1.value = "make_image_crate"
    key1.desc = "tool that created this output"
    cr.add_key(key1)

    key2 = pycrates.CrateKey()
    key2.name = "DATE"
    key2.value = time.strftime(_fits_time_fmt)
    key2.desc = "Date and time of file creation"
    cr.add_key(key2)

    return cr


# Used by write_columns
#
__kernel_mapping = {
    "text": "text/simple",
    "simple": "text/simple",
    "dtf": "text/dtf",
    "raw": "text/raw",
    "fits": None
}


def write_columns(fname, *args, **kwargs):
    """Write out arrays as columns in the file fname.

    After the filename come the array values (these can be Python
    arrays or numpy arrays, and include strings). Python arrays should
    contain the same data type for each element; if this does not hold
    then there is no guarantee as to the behavior of this routine. If
    they are not the same length then the smaller arrays will be
    padded with the null value for that column type (0 for numeric
    columns and 'IND' for string columns).

    The supported keyword arguments, along with their default values,
    are:

     colnames=None

      When the input is a dictionary or structured array, this is
      a list of columns to use (and so should match or be a subset of
      the column names/keys in the input).

      Otherwise, if left as None, then the columns will be called 'col1',
      'col2', ..., 'coln' (where n is the number of arrays sent in).
      If given, it should be an array with at least n elements which
      are used as the column names (extra values are ignored).

     format='text'

      The format for the output file: the supported values are 'text',
      'fits', 'simple' (same as 'text', 'dtf', and 'raw'. See 'ahelp
      dmascii' for information on the various ASCII formats.

     sep=' '

      The column separator. Only used when type is not 'fits'.

     comment='#'

      The comment character (i.e. this character starts comment
      lines). Only used when type is not 'fits'.

    clobber=True (default) or False

      If the output file exists then it will be overwritten (when True) or
      cause the routine to raise an IOError (when set to False).

    Examples:

      write_columns('cols.dat', a, b)
      write_columns('cols.dat', a, b, colnames=['x','y'])
      write_columns('cols.dat', a, b, colnames=['x','y'], format='fits')
      write_columns('cols.dat', a, b, colnames=['x','y'], clobber=False)

      d = { "x": a, "y": b, "z": a + b }
      write_columns('cols.dat', d)
      write_columns('cols.dat', d, colnames=["x", "y"])

    """

    # We repeat some usage checks that make_table_crate() does so that
    # we give a more user-friendly error message than if it came from a
    # routine the user did not directly call.
    #
    if len(args) == 0:
        raise TypeError("write_columns() takes at least two arguments (1 given)")

    fmt = kwargs.pop("format", "text")

    try:
        kernel = __kernel_mapping[fmt]
    except KeyError:
        raise TypeError("format argument sent '{0}', must be one of {1}".format(fmt, ", ".join(__kernel_mapping.keys())))

    colsep = kwargs.pop("sep", " ")
    if len(colsep) != 1:
        raise TypeError("sep argument must be a single character, sent '{0}'".format(colsep))

    comment = kwargs.pop("comment", "#")
    if len(comment) != 1:
        raise TypeError("comment argument must be a single character, sent '{0}'".format(comment))

    clobber = kwargs.pop("clobber", True)

    for k in kwargs.keys():
        if k != "colnames":
            raise TypeError("write_columns() got an unexpected keyword argument '{}'".format(k))

    cr = make_table_crate(*args, **kwargs)

    oname = fname
    if kernel is not None:
        # In CIAO 4.4 release 1, keywords are written out for ASCII formats
        # when the comment option is set. Since most people will use the
        # default comment option we avoid this by only including the comment
        # value if not '#'; this is a hack. Is this still true?
        if comment == '#':
            oname += "[opt kernel={0},sep='{1}']".format(kernel, colsep)
        else:
            oname += "[opt kernel={0},sep='{1}',comment={2}]".format(kernel, colsep, comment)

    cr.write(oname, clobber=clobber)


def write_arrays(filename, args, fields=None, sep=' ', comment='#',
                 clobber=False, linebreak='\n', format='%g'):
    """Write a list of arrays to an ASCII file.

    Parameters
    ----------
    filename : str
        The name of the file to create.
    args : list
        The list of arrays, where each element represents a column of data.
    fields : list of str, optional
        The column names to use. If not given then the names 'col1', 'col2',
        ... will be used.
    sep : str, optional
        The character used to separate columns.
    comment : str, optional
        The comment character (at the start of a line).
    clobber : bool, optional
        Should the file be over-written if it already exists?
    linebreak : str, optional
        The characters used to indicate a line break.
    format : str, optional
        The format string used to convert each column element into a
        string.

    See Also
    --------
    write_columns

    Examples
    --------

    >>> a = [0,1,2,3,4]
    >>> b = [1,2,3,4,5]
    >>> c = [1,2,4,8,16]
    >>> write_arrays('out.dat', [a,b,c], fields=['a','b','c'], clobber=True)
    """
    if os.path.isfile(filename) and not clobber:
        raise IOError("file '{0}' exists and clobber is not set".format(filename))

    if not np.iterable(args) or len(args) == 0:
        raise IOError('please supply array(s) to write to file')

    if not np.iterable(args[0]):
        raise IOError('please supply array(s) to write to file')

    size = len(args[0])
    for arg in args:
        if not np.iterable(arg):
            raise IOError('please supply array(s) to write to file')
        elif len(arg) != size:
            raise IOError('not all arrays are of equal length')

    args = np.column_stack(np.asarray(args))

    with open(filename, 'w') as f:
        if fields is not None:
            f.write(comment + sep.join(fields) + linebreak)

        lines = []
        for arg in args:
            line = [format.format(elem) for elem in arg]
            lines.append(sep.join(line))

        f.write(linebreak.join(lines))

        # add a newline at end
        f.write(linebreak)


_image_scalings = {
    "none": None,
    "arcsinh": np.arcsinh,
    "asinh": np.arcsinh,
    "sqrt": np.sqrt,
    "square": np.square,
    "log10": np.log10,
    "log": np.log
}


def scale_image_crate(cr, scaling="arcsinh"):
    """Apply a scaling function to the pixel values in the
    crate. The currently-supported routines are

       none              - no scaling (does nothing)
       arcsinh or asinh
       sqrt
       square
       log10
       log        [natural logarithm]

    This routine modifies the data in cr; it does NOT return
    a copy of the crate.
    """

    if not isinstance(cr, pycrates.IMAGECrate):
        raise ValueError("First argument must be an image crate.")

    # Short cut
    if scaling == "none":
        return

    ivals = pycrates.copy_piximgvals(cr)

    try:
        sfunc = _image_scalings[scaling]

    except KeyError:
        raise ValueError("Invalid scaling value of '{0}': must be one of\n  {1}".format(
            scaling, " ".join(_image_scalings.keys())))

    pycrates.set_piximgvals(cr, sfunc(ivals))


_image_smooth = {
    "none": None,
    "gauss": sm.gsmooth,
    "boxcar": sm.bsmooth,
    "tophat": sm.tsmooth,
    "image": sm.ismooth,
    "file": sm.fsmooth
}


def smooth_image_crate(cr, stype, *args, **kwargs):
    """Smooth the pixel values in the crate. The available options
    are controlled by the type argument which accepts:

    type    smoothing  args/kwargs
    none    no scaling
    gauss   gaussian   sigma, nsigma=5
    boxcar  box-car    radius
    tophat  top-hat    radius
    file    by file    file name, norm=True, origin=None
    image   by image   numpy array, norm=True, origin=None

    The remaining arguments are those that you would supply to the
    gsmooth(), bsmooth(), tsmooth(), fsmooth() or ismooth()
    routines from the ciao_contrib.smooth module.

    This routine modifies the data in cr; it does NOT return
    a copy of the crate.
    """

    if not isinstance(cr, pycrates.IMAGECrate):
        raise ValueError("First argument must be an image crate.")

    # Short cut
    if stype == "none":
        return

    ivals = pycrates.copy_piximgvals(cr)

    try:
        sfunc = _image_smooth[stype]

    except KeyError:
        raise ValueError("Invalid smooth value of '{0}': must be one of\n  {1}".format(
            stype, " ".join(_image_smooth.keys())))

    nvals = sfunc(ivals, *args, **kwargs)
    pycrates.set_piximgvals(cr, nvals)


# Coordinate-conversion code

class SimpleCoordTransform:
    """Support simple coordinate transforms between
    image coordinate systems (e.g. logical to physical to world).

    It is only likely to support CIAO-produced image files.

    """

    # The mapping between Transform class name and the user-friendly
    # label.
    _coord_mapping = {
        # In CIAO 4.6 we now have the WCSTransform class; the WCSTANTransform
        # version is still supported for now but is assumed to be deprecated.
        'WCSTANTransform': 'world',
        'WCSTransform': 'world',
        'LINEAR2DTransform': 'physical'
    }

    # The mapping between user-supplied coordinate-system name and
    # the internal names (case insensitive).
    _sys_mapping = {
        "world": "world", "eqpos": "world",
        "physical": "physical", "sky": "physical",
        "logical": "logical", "image": "logical"
    }

    def __init__(self, inval):
        """Set up the coordinate transforms for the image
        in inval, which can either be an IMAGECrate or a
        filename.
        """

        if hasattr(inval, "get_axisnames") and hasattr(inval, "get_transform"):
            crate = inval
            fname = crate.get_filename()
        else:
            crate = pycrates.IMAGECrate(inval, mode="r")
            fname = inval

        axes = crate.get_axisnames()
        if axes is None or len(axes) == 0:
            raise ValueError("No axes found in {0}".format(fname))

        # get_transform is case-sensitive
        self.axes = [n.split('(')[0] for n in axes]
        self.transforms = {}
        for n in self.axes:
            try:
                tr = crate.get_transform(n).copy()
            except ValueError:
                raise ValueError("Unable to find the transformation for {0}".format(n))

            try:
                ctype = self._coord_mapping[tr.get_className()]
            except KeyError:
                continue

            try:
                self.transforms[ctype]
                raise ValueError("Found multiple {0} coordinate systems in {1}".format(ctype, fname))

            except KeyError:
                self.transforms[ctype] = {"name": n, "type": ctype,
                                          "transform": tr}

        try:
            world = self.transforms["world"]
            physical = self.transforms["physical"]

        except KeyError, ke:
            raise ValueError("Unable to find the {0} system in {1}".format(ke.args[0], fname))

        # self.conversions[fromsys][tosys]
        #
        self.conversions = {}
        self.conversions["world"] = {
            "world": None,
            "physical": {"direction": "invert", "transforms": [world]},
            "logical": {"direction": "invert", "transforms": [world, physical]}
        }
        self.conversions["physical"] = {
            "world": {"direction": "apply", "transforms": [world]},
            "physical": None,
            "logical": {"direction": "invert", "transforms": [physical]}
        }
        self.conversions["logical"] = {
            "world": {"direction": "apply", "transforms": [physical, world]},
            "physical": {"direction": "apply", "transforms": [physical]},
            "logical": None
        }

    def _validate_coordsys(self, cname):
        "Raise a ValueError if cname is invalid, returns the internal name"

        try:
            return self._sys_mapping[cname.lower()]

        except KeyError:
            raise ValueError("Invalid coordinate-system name: " +
                             "{0}".format(cname))

    # Unfortunately I did not note why I only do this point-by-point rather
    # than send in all the points at once. There have been issues with checking
    # coordinate dimensionality, that have been fixed in CIAO 4.5, that
    # may mean that we can go back to just handling vectors, but for now
    # keep as is.
    #
    def _convert(self, direction, mappings, x, y):
        """Convert a single (x,y) using the direction and mappings."""

        pt = [[x, y]]
        for trinfo in mappings:
            pt = getattr(trinfo["transform"], direction)(pt)

        return pt[0, :]

    def convert(self, fromsys, tosys, x, y):
        """Convert the array of coordinates in x,y from the fromsys
        coordinate system to tosys; the accepted values are:

           world, eqpos
           physical, sky
           logical, image

        The return value is a pair of values. Note that x and y must
        be numpy arrays or something that can be passed through np.asarray,
        and can be multidimensional as long as the shapes are the same.

        """

        v3("Coordinate conversion of {},{} from {} to {}".format(x, y, fromsys, tosys))
        fsys = self._validate_coordsys(fromsys)
        tsys = self._validate_coordsys(tosys)
        try:
            cinfo = self.conversions[fsys][tsys]

        except KeyError, ke:
            raise ValueError("Unable to find transforms for {0} (unexpected)".format(ke.args[0]))

        v3(" . conversion info = {}".format(cinfo))
        if cinfo is None:
            return (np.asarray(x), np.asarray(y))

        d = cinfo["direction"]
        trs = cinfo["transforms"]

        xin = np.asarray(x)
        yin = np.asarray(y)

        # what was I going to do with this?
        # otype = np.common_type(xin, yin)

        # Could presumably be a lot more clever with array manipulation here
        b = np.broadcast(xin, yin)
        v3(" . to convert {}".format(b))
        out = np.asarray([self._convert(d, trs, u, v) for (u, v) in b])
        v3(" . answer = {}".format(out))
        xout = out[:, 0].copy().reshape(b.shape)
        yout = out[:, 1].copy().reshape(b.shape)
        return xout, yout


def read_ds9_contours(fname, coords=None, fromsys=None, tosys=None):
    """Read in a DS9 *.con file and return the
    x and y line segments as an array of arrays.

      (xall, yall) = read_ds9_contours("ds9.con")
      print("Number of segments: {0}".format(len(xall)))
      for i in range(0, len(xall)):
          add_line(xall[i], yall[i], ["color", "green"])

    The coords, fromsys, and tosys arguments are used to convert
    the coordinates to a different system: coords is either the
    name of an image or a crate that contains the desired coordinate
    systems, fromsys is the name of the coordinate system used to
    store the contours and tosys the name to convert to. The valid
    values for fromsys and tosys are:

      world, eqpos
      physical, sky
      logical, image

    See the add_ds9_contours routine in chips_contrib.ds9
    for a way to automate this.

    """

    xall = []
    yall = []
    x = []
    y = []

    if None in [coords, fromsys, tosys]:
        tr = None
    else:
        tr = SimpleCoordTransform(coords)

    with open(fname, "r") as fh:
        for l in fh.readlines():
            l = l.strip()
            if l == '':
                if len(x) == 0:
                    raise ValueError("Unexpected blank line: read in x=\n{0}\n".format(xall))

                if tr is None:
                    xo = np.asarray(x)
                    yo = np.asarray(y)
                else:
                    (xo, yo) = tr.convert(fromsys, tosys, x, y)

                xall.append(xo)
                yall.append(yo)

                x = []
                y = []

            else:
                toks = l.split()
                if len(toks) != 2:
                    raise ValueError("Unable to parse line: '{0}'".format(l))

                x.append(float(toks[0]))
                y.append(float(toks[1]))

    if len(x) != 0:
        if tr is None:
            xo = np.asarray(x)
            yo = np.asarray(y)
        else:
            (xo, yo) = tr.convert(fromsys, tosys, x, y)

        xall.append(xo)
        yall.append(yo)

    if len(xall) == 0:
        raise ValueError("No data read in from: {0}".format(fname))

    return (xall, yall)

# End
